#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.ipc.win_mutex
#
# Description:  mutex for windows os
#
# #############################################################################
import os

if os.name == 'win32':
    from win32event import CreateMutex
    from win32event import WaitForSingleObject
    from win32event import ReleaseMutex
    from win32api import GetLastError
    from winerror import ERROR_ALREADY_EXISTS


#                      76543210
TIMEOUT_INFINITE   = 0xffffffff  # noqa: E221
MTX_WAIT_ABANDONED = 0x00000080  # noqa: E221
MTX_WAIT_OBJECT_O  = 0x00000000  # noqa: E221
MTX_WAIT_TIMEOUT   = 0x00000102  # noqa: E221
MTX_WAIT_FAILED    = 0xffffffff  # noqa: E221


class IPCMutexError(Exception):
    """ Errors Specific to this module """


class IPCMutex(object):
    """ Basic wrapper around windows Named mutex object
        due to legacy reasons there's still the initial acquire mode,
        which is intended for ensuring, that a certain executable is
        called only once. (will_own=1)
        In this case the calling process insists to own the mutex and
        fails otherwise.
        In the other contexts the first process will create the mutex.
    """

    def __init__(self, name, will_own=1, lockdir=None):
        self.name = name
        self.will_own = will_own
        if not will_own:
            self.handle = CreateMutex(None, self.will_own, self.name)
        else:
            self.handle = None

    def acquire(self, timeout=TIMEOUT_INFINITE):
        """ gets mutex
            due to legacy reasons there's still the initial acquire mode,
            which is intended for ensuring, that a certain executable is
            called only once. (will_own=1)
        """
        if self.will_own:
            self.handle = CreateMutex(None, self.will_own, self.name)
            err = GetLastError()
            if err == ERROR_ALREADY_EXISTS:
                return False
            else:
                return self
        rslt = WaitForSingleObject(self.handle, timeout)
        if rslt == MTX_WAIT_OBJECT_O:
            return self
        elif rslt == MTX_WAIT_TIMEOUT:
            return False
        raise IPCMutexError("got got return code %08x" % rslt)

    def release(self):
        """ releases a mutex. Raises exception if error """
        # TODO: could fetch enhanced error info with GetLastError()
        rslt = ReleaseMutex(self.handle)
        if rslt == 0:
            raise IPCMutexError("could not release Mutex.")



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.tests.tst_gitlab
#
# Description:  test cases for mytb.gitlab.check_ci_cfg
#
# #############################################################################
# python std modules
import os
import unittest

# first party modules
import yaml
from ddt import data, ddt

# mytb modules
from mytb.gitlab.check_ci_cfg import check_cfg_file
from mytb.tempfile import mktemp_fname
from mytb.tests.gitlab.cicfg_patterns import fail_patterns, pass_patterns


@ddt
class GitlabCfg(unittest.TestCase):

    def setUp(self):
        self.tmpfiles = []

    def tearDown(self):
        for fname in self.tmpfiles:
            if os.path.isfile(fname):
                os.unlink(fname)

    def test_can_read_file(self):
        """ can read gitlab-ci-cfg files """
        fname = mktemp_fname()
        self.tmpfiles.append(fname)
        with open(fname, "w") as fout:
            fout.write(yaml.dump(dict(before_script="ls")))

    @data(*pass_patterns)
    def test_pattern_passes(self, pattern):
        """ gitlab-ci-cfg is correct """
        self.assertTrue(check_cfg_file(text=pattern.value))

    @data(*fail_patterns)
    def test_pattern_fails(self, pattern):
        """ gitlab-ci-cfg has errors """
        self.assertFalse(check_cfg_file(text=pattern.value))

#     def test_get_rls_mutex(self):
#         """ can acquire and release a mutex """
#         mutex = Mutex("toto")
#         status = mutex.status
#         print("MST", status)
#         self.assertFalse(mutex.status)
#         mutex.acquire(timeout=1)
#         self.assertTrue(mutex.status)
#         mutex.release()
#         self.assertFalse(mutex.status)
#     def test_ctx_get_rls_mutex(self):
#         """ can acquire and release with context manager """
#         with Mutex("toto", 1) as mutex:
#             pass
#             #self.assertTrue(mutex.status)
#         #self.assertFalse(mutex.status)
#         # can acquire mutex without timeout
#         mutex = Mutex("toto", 1)
#
#
# if __name__ == "__main__":
#     import sys
#     args = sys.argv[1:]
#     if args and args[0] == "test":
#         print("TST")
#         suite = unittest.TestSuite()
#         suite.addTest(IPCTestCase("test_get_rls_mutex"))
#         runner = unittest.TextTestRunner(verbosity=2, stream=sys.stdout)
#         runner.run(suite)



from unidecode import unidecode


def str2bool(val):
    """ converts a string to a boolean value
        if a non string value is passed
        then following types will just be converted to Bool:
            int, float, None

        other types will raise an exception

        inspired by
        https://stackoverflow.com/questions/15008758/
        parsing-boolean-values-with-argparse
    """
    if val is None:
        return None
    if isinstance(val, (bool, int, float)):
        return bool(val)
    if not hasattr(val, "lower"):
        pass
    elif val.lower() in ("yes", "oui", "ja", "true", "t", "y", "1"):
        return True
    elif val.lower() in ("no", "non", "nein", "false", "f", "n", "0", ""):
        return False

    raise ValueError("can't convert %r to bool" % val)


def strip_accents(astring):
    """
    Accept a unicode string, and returns a string (type str)
    without any diacritical marks.
    """
    rslt = unidecode(astring)
    return rslt



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.tests.test_argparse
#
# Description:  unit tests for argparser
#
# #############################################################################

import argparse
import unittest


class ArgParseTestCase(unittest.TestCase):
    def test_mkparser(self):
        """ 1 1 2 """
        import mytb.argparse
        parser = mytb.argparse.mk_parser()

        #  find all expected keys in parser
        options = parser.parse_args([])
        self.assertEqual(
            set(vars(options).keys()),
            set(['app_name', 'log_config']))
        self.assertEqual(
            options.log_config,
            'mytb.logging.configs.console:level=WARNING',
            )

        options = parser.parse_args(['-L', ''])
        self.assertEqual(
            options.log_config,
            '',
            )

        #  -L switch works
        options = parser.parse_args(['-L', 'tido'])
        self.assertEqual(options.log_config, 'tido')

        #  --log-config switch works
        options = parser.parse_args(['--log-config', 'todo'])
        self.assertEqual(options.log_config, 'todo')

        #  can override long switch
        mytb.argparse.LONG_LOG_SWITCH = '--lock-cfg'
        parser = mytb.argparse.mk_parser()
        options = parser.parse_args(['--lock-cfg', 'tata'])
        self.assertEqual(options.lock_cfg, 'tata')
        print(options)


def test_add_bool_arg():
    from mytb.argparse import add_bool_argument
    parser = argparse.ArgumentParser()
    add_bool_argument(parser, "-s")
    add_bool_argument(parser, "--long")
    add_bool_argument(parser, "-b", "--boolflag")
    add_bool_argument(
        parser, "-d", "--dataflag",
        help="this is a help text",
        )

    #
    args = ["-s", "--long", "0"]
    options = parser.parse_args(args)
    print("options", options)
    assert options.s is True
    assert options.long is False
    assert options.boolflag is None

    for val in ["1", "true", "yes"]:
        print("options", options)
        args = ["-s", "--long", "0", "-b", val]
        options = parser.parse_args(args)
        assert options.boolflag is True

    for val in ["0", "f", "no"]:
        print("options", val, options)
        args = ["-s", "--long", "0", "-b", val]
        options = parser.parse_args(args)
        assert options.boolflag is False



from datetime import datetime
from datetime import timedelta

import pytz
import tzlocal

import mytb.datetime

from mytb.datetime import DateTime
# from mytb.datetime import Date
# from mytb.datetime import Time
# from mytb.datetime import fname_to_time
from mytb.datetime import to_timestamp


def check_diff(t1, t2, maxdelta=0.01):
    """ checks whether two time values are sufficiently close """
    absdiff = abs((t1 - t2).total_seconds())
    print(absdiff)
    return absdiff < maxdelta


def test_to_timestamp():
    """ we can convert to epoch seconds
    """
    # check that T_EPOCH is correct
    assert to_timestamp(mytb.datetime.T_EPOCH) == 0

    # check that 1970-01-01 00:00:00 UTC is well 0
    t0 = datetime(1970, 1, 1, tzinfo=pytz.utc)
    assert to_timestamp(t0) == 0

    # check that today can be converted back and forth via utc
    mytz = tzlocal.get_localzone()
    now = datetime.now(mytz)
    ts = to_timestamp(now)

    # check that today can be converted back and forth via local
    utcfromtimestamp = datetime.utcfromtimestamp
    back_utc = pytz.utc.localize(utcfromtimestamp(ts))
    assert back_utc == now

    back_local = mytz.localize(datetime.fromtimestamp(ts))
    assert back_local == now

    # works even after year 2038
    y2040 = datetime(2040, 2, 29, 17, 48, 12, tzinfo=mytz)
    ts = to_timestamp(y2040)
    back_utc = pytz.utc.localize(utcfromtimestamp(ts))
    assert back_utc == y2040


def test_datetime_parse():
    """ we can parse our strings """

    # helper vars
    utc = pytz.utc

    # test now parsing
    now = datetime.now(utc)
    for nowval in (None, '', 0, 'now'):
        mynow = DateTime.strptime(nowval)
        delta = (mynow - now).total_seconds()
        assert delta < 0.01

    # test dflt strptime
    adate = DateTime.strptime("2017-01-01", "%Y-%m-%d", tzinfo=None)
    assert adate == datetime(2017, 1, 1)

    # same but with zime zones
    adate = DateTime.strptime("2017-01-01", "%Y-%m-%d", utc)
    assert adate == utc.localize(datetime(2017, 1, 1))


def test_range():
    now = DateTime.strptime('now')
    t_from, t_to = DateTime.parse_range()
    one_day = timedelta(1)
    assert check_diff(t_from, now - one_day)
    assert check_diff(t_to, now)



"""
Helper to get info about memory consumption
"""

import logging
import os

import psutil


logger = logging.getLogger(__name__)


def get_rss():
    pid = psutil.Process(os.getpid())
    return pid.memory_info().rss


def log_rss():
    logger.debug("RSS: %d", get_rss())



#!/usr/bin/env python

# ############################################################################
# Copyright  : (C) 2015-2021 by MHComm. All rights reserved
#
# Name       :  mytb.pickle
"""
  Summary    : pickle / unpickle related helpers

__author__    = "Klaus Foerster"
__email__     = "info@mhcomm.fr"


Current implementation is just a dummy for the mhflt tool.

robust pickling / unpickling should be implemented lateron
"""
# #############################################################################
import pickle

from io import BytesIO


def robust_unpickler(fin):
    """
    an unpickler that tries to be more robust.

    At the moment this is just a dummy
    """
    return pickle.Unpickler(fin)


def unpickle_robust(bytestr):
    """ robust unpickle of one byte string """
    fin = BytesIO(bytestr)
    unpickler = robust_unpickler(fin)
    return unpickler.load()



import asyncio
import sys

# Not tested for pre 3.4 versions
assert sys.version_info >= (3, 4)

patched = []  # keeps track of what has been patched


def patch():
    """
    patches pre 3.7 asyncios with some hacky but 'compatible' back ports
    """

    if not hasattr(asyncio, "run"):
        # asyncio.run function has been added to asyncio in
        # Python 3.7 on a provisional basis.
        #

        def run(coro):
            loop = asyncio.get_event_loop()
            loop.create_task(coro)
            loop.run_forever()

        asyncio.run = run
        patched.append("asyncio.run")

    if not hasattr(asyncio, "all_tasks"):
        # from python 3.7 on asyncio.all_tasks replaces asyncio.Task.all_tasks,
        # which will be removed in Python 3.9
        asyncio.all_tasks = asyncio.Task.all_tasks
        patched.append("asyncio.all_tasks")



from itertools import count

from mytb.lru_cache import lru_cache

already_called = {}

ctr = count()
ctr2 = count()


def mk_signature(args, kwargs):
    sig = tuple(args), tuple(kwargs.items())
    print("sig(%s, %s) => %s" % (args, kwargs, sig))
    return sig


@lru_cache(maxsize=128)
def tstfunc(*args, **kwargs):
    """ a function, that returns a unique result for every call """
    rslt = mk_signature(args, kwargs), next(ctr)
    print("Func(%s, %s) => %s" % (args, kwargs, rslt))
    return rslt


@lru_cache(maxsize=4)
def tstfunc2(*args, **kwargs):
    """ a function, that returns a unique result for every call """
    rslt = mk_signature(args, kwargs), next(ctr2)
    print("Func(%s, %s) => %s" % (args, kwargs, rslt))
    return rslt


def test_cache():

    params = (
        ((1, ), {}),
        ((2, ), {}),
        ((1, 2, ), {}),
        (('1', ), {}),
        ((), {'a': 'val'}),
        )

    results = set()
    for args, kwargs in params:
        signature = mk_signature(args, kwargs)
        rslt_signature, rslt = tstfunc(*args, **kwargs)
        print("CALL1: ", signature, rslt_signature, rslt)
        assert signature == rslt_signature

        rslt_signature, rslt2 = tstfunc(*args, **kwargs)
        print("CALL2: ", signature, rslt_signature, rslt2)
        assert signature == rslt_signature
        assert rslt == rslt2
        results.add(rslt)

    results = list(results)
    print("RSLTS", results)

    assert len(results) == len(params)


def test_cache_size():
    """ show, that cache expires if and only if too many results
        are cached
    """
    call1_rslts = []
    for val in range(4):
        rslt_signature, rslt = tstfunc2(val)
        call1_rslts.append((rslt_signature, rslt))

    for val, call1_rslt in zip(range(4), call1_rslts):
        rslt_signature2, rslt2 = tstfunc2(val)
        rslt_signature, rslt = call1_rslt
        assert rslt_signature == rslt_signature2
        assert rslt == rslt2

    call1_rslts = []
    for val in range(5):
        rslt_signature, rslt = tstfunc2(val)
        call1_rslts.append((rslt_signature, rslt))

    for val, call1_rslt in zip(range(5), call1_rslts):
        rslt_signature2, rslt2 = tstfunc2(val)
        rslt_signature, rslt = call1_rslt
        assert rslt_signature == rslt_signature2
        assert rslt + 5 == rslt2



#!/usr/bin/env python

# ############################################################################
# Copyright  : (C) 2014-2021 by MHComm. All rights reserved
#
# Name       : mytb.file.suffix_file
"""
  Summary    :  opens files depending on suffix with normal open gzip or bzip

__author__ = "Klaus Foerster"
__email__ = "info@mhcomm.fr"

"""
# #############################################################################


class SuffixFile(object):
    """ Fileobject, that will inspect its suffix to decide how to open
    """
    # TODO: Implement suffixes
    @classmethod
    def open(cls, fname, mode="r", **kwargs):
        """ opens file with respective compression / decompression """
        return open(fname, mode, **kwargs)



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2021 by Teledomic.eu All rights reserved
#
# Name:         mytb.parse
#
# Description:  helpers for parsing
#
# #############################################################################
import re

from collections import OrderedDict


def parse_paramstr(argstr, sep=None, sep_re=None):
    """ parse function to parse name1=val2:name2=val2 . . .

        no separator allowed in value strings or keys

        :param argstr: the string to be parsed
        :param sep: if set use to split assigments
        :param sep_re: if set use regex or regex string as separator
        :returns: OrderedDict of keys, values
    """
    assert not(sep and sep_re)  # only one of the params should be set
    if not (sep or sep_re):
        sep = ":"
    if sep:
        assigns = argstr.split(sep)
    else:
        assigns = re.split(sep_re, argstr)

    rslt = OrderedDict()
    for kv in assigns:
        key, val = kv.split('=', 1)
        rslt[key] = val
    return rslt



# #############################################################################
# Copyright : (C) 2019 by Teledomic.eu All rights reserved
#
# Name:         mytb.ipython
#
# Description:  helper to create an interactive python shell
#
# Ipython will be used if installed
#
# #############################################################################


try:
    import IPython
    HAVE_IPYTHON = True
except Exception:
    HAVE_IPYTHON = False
    raise


def simple_shell(ctx):
    """ basic interactive shell if ipython is not found
    """
    try:
        import readline  # noqa: F401
    except Exception:
        print("readline not found")
    import code

    shell = code.InteractiveConsole(ctx)
    shell.interact()


def start_shell(ctx=None):
    """ starts an interactive python shell
    """
    ctx = ctx if ctx is not None else {}
    if HAVE_IPYTHON:
        IPython.start_ipython(argv=[], user_ns=ctx)
        return
    else:
        print("ipython not installed")
        simple_shell(ctx)



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.ipc.mutex
#
# Description:  crossplatform IPC mutex (windows / linux)
#
# #############################################################################
from mytb.os import is_posix, is_windows

if is_posix():
    from .posix_mutex import PosixMutex as IPCMutex
elif is_windows():
    from .win_mutex import WinMutex as IPCMutex
else:
    raise NotImplementedError("mutex for this platform not implemented")

_mutexes = dict()


def Mutex(self, name, lockdir=None):
    """ factory returning an OS specific mutex """
    mutex = _mutexes.get(name) or IPCMutex(name, lockdir=lockdir)
    _mutexes[name] = mutex
    return mutex



#!/usr/bin/env python

# ############################################################################
# Copyright  : (C) 2019 by MHComm. All rights reserved
#
# Name       : mytb.debug
"""
  Summary    :  Helpers for debugging

__email__ = "info@teledomic.eu"
"""
# #############################################################################

import sys
import traceback


def get_debugger_mod():
    """ import and gets a debugger module.
        It will try to import ipbd and if it is not installed the default pdb
    """
    try:
        import ipdb as dbg_mod
    except ImportError:
        import pdb as dbg_mod
    return dbg_mod


def pdb_debug_hook(type_, value, tb):
    """ a debug hook function, that can for example be assigned to
        sys.excepthook in order to start a debugger whenever
        an error occured.

        It also displays a small help text for people who're not
        that used to the python debugger commands.

        This hook will only be activated if run in a terminal
    """

    if not sys.stderr.isatty():  # do not start debugger if no console
        sys.__excepthook__(type_, value, tb)
        return

    dbgr_mod = get_debugger_mod()

    traceback.print_exception(type_, value, tb)
    if dbgr_mod.__name__ == 'pdb':
        print(
            "you can install ipdb to have"
            " a python debugger with slightly better"
            " interactive behaviour (tab-completion)\n")
    print(
        "Press:\n"
        "'w' to get a the stack frame,\n"
        "'u' to move up the stack frame\n"
        "'d' to move down the stack frame\n"
        "'q' to quit\n"
        "'h' for help\n"
        )

    dbgr_mod.pm()



import sys

import pytest


@pytest.mark.skipif(
    sys.version_info < (3, 4),
    reason="requires python3.4+")
def test_aio_compat():
    """ can patch old asyncios """
    import asyncio
    import mytb.aio.compat
    mytb.aio.compat.patch()

    assert asyncio.run
    assert asyncio.all_tasks

    patched = mytb.aio.compat.patched
    if sys.version_info < (3, 7):
        assert "asyncio.run" in patched
        assert "asyncio.all_tasks" in patched



import os

from mytb.exceptions import MyTBError


def robust_makedirs(path):
    """ create a directory in a robust race safe manner if not
        already existing.
        Good for multiprocessing / threading or cases where
        multiple actors might create a directory
    """
    if not os.path.isdir(path):
        try:
            os.makedirs(path)
        except Exception:
            if not os.path.isdir(path):
                if os.path.isfile(path):
                    raise MyTBError("path %r is not a directory" % path)
                else:
                    raise



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.gitlab.check_ci_cfg
#
"""
    validates a gitlab-ci config file rather superficially,
    but still better than no syntax check at all. might be implemented as a
    pre-commit trigger
"""
#
# #############################################################################
# python std modules
import argparse
import re
import sys

# first party modules
import yaml


rex = re.compile  # compile a regex


def print_nothing(*args, **kwargs):
    pass


trace = print_nothing  # for tracing in verbose mode


class Entry(object):
    def init(self, name_rex, *args, **kwargs):
        self.name = name_rex


def match(key, pattern):
    if hasattr(pattern, 'match'):
        return pattern.match(key)
    else:
        return key == pattern


# attemp of describing a very superficial syntax tree
tree = (
    (
        'cache', (
            ("key",  True),
            ("paths", list),
            ),
        ),
    ('stages', list),
    ('before_script', list),
    (
        rex(r".*"), (
            ('allow_failure', bool),
            ('image', str),
            ('stage', str),
            ('script', list),
            ('tags', list),
            )),
)


def check_cfg(cfg, tree, parent=''):
    """ reursiveley check a cfg struct against a simple syntax tree """
    if type(cfg) is dict:
        for key, value in cfg.items():
            trace("\nKEY", key)
            for tree_entry in tree:
                pattern, subtree = tree_entry
                if match(key, pattern):
                    trace("tree entry", tree_entry)
                    trace("SUBTREE", subtree)
                    treetype = type(subtree)
                    if subtree is True:
                        break
                    elif subtree is str:
                        if type(value) is str:
                            break
                        print("%s.%s must be string" % (parent, key))
                        return False
                    elif subtree is bool:
                        if type(value) is bool:
                            break
                        print("%s.%s must be bool" % (parent, key))
                        return False
                    elif subtree is list:
                        if type(value) is list:
                            break
                        print("%s.%s must be list" % (parent, key))
                        return False
                    elif treetype in (tuple,):
                        if parent:
                            sub_parent = parent + '.' + key
                        else:
                            sub_parent = key
                        result = check_cfg(value, subtree, parent=sub_parent)
                        if not result:
                            return result
                        break
                    else:
                        print("NOT IMPLEMENTED: %s.%s" % (parent, key))
                        return False
            else:
                print("key %s.%s doesn't match" % (parent, key))
                return False
            trace("found")
    return True


def check_cfg_file(fname=None, text=None):
    """ checks whether a gitlabci config file seems to be correct
        The checks are not very extensive, but capture some common
        cases.
    """
    if (sum((1 if val else 0) for val in (fname, text is not None))
            != 1):
        raise Exception("must pass either fname or text")

    if fname:
        with open(fname) as fin:
            text = fin.read()
    try:
        cfg = yaml.load(text)
    except (yaml.parser.ParserError,
            yaml.scanner.ScannerError) as exc:
        print("yaml syntax error")
        print(exc)
        return False

    print("yaml syntax is correct")
    rslt = check_cfg(cfg, tree)
    print("Tree check = ", rslt)
    return rslt


def mk_parser():
    parser = argparse.ArgumentParser(
        description="checks syntax of gitlab ci config file")
    parser.add_argument(
        "--verbose", "-v", action="store_true",
        help="verbose display",
        )
    parser.add_argument(
        "fname", nargs="?", default='.gitlab-ci.yml',
        help="filename to check (default=%(default)s",
        )
    return parser


def main():
    global trace
    args = sys.argv[1:]
    parser = mk_parser()
    options = parser.parse_args(args)
    if options.verbose:
        trace = print
    fname = options.fname
    ok = check_cfg_file(fname)
    sys.exit(0 if ok else -1)



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.ipc.posix_mutex
#
# Description:  mutex for posix OS implemented via a lock file
#
# #############################################################################

import os
import fcntl
from tempfile import tempdir


class IPCMutex(object):
    def __init__(self, name, will_own=None, lockdir=tempdir):
        self.name = name
        self.lockdir = lockdir
        self.lockfile = None

    def acquire(self):
        """ acquires a mutex """
        lockfile_name = os.path.join(self.lockdir, self.name)
        cur_umask = os.umask(0)
        lockfile = open(lockfile_name, 'w')
        os.umask(cur_umask)
        try:
            fcntl.lockf(lockfile, fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.lockfile = lockfile
            return self
        except IOError:
            return False

    # will be automatically release if process ends
    def release(self):
        raise NotImplementedError("cannot release this mutex at the moment")



import os

import pytest

from mytb.exceptions import MyTBError
from mytb.path import robust_makedirs


def test_robust_makedirs(tmp_path, monkeypatch):

    # create new dir and check no fail if existing
    dir1 = str(tmp_path / "d1" / "d2")
    robust_makedirs(dir1)
    assert os.path.isdir(dir1), "can create non existing dir"
    robust_makedirs(dir1)
    assert os.path.isdir(dir1), "pass on existing dir"

    dir2 = os.path.join(dir1, "f3")
    dir3 = os.path.join(dir2, "d4")
    with open(dir2, "w"):  # create empty file
        pass

    # can't create a directory on top of an existing file
    with pytest.raises(MyTBError):
        robust_makedirs(dir2)

    # can't create a directory below a file
    with pytest.raises(OSError):
        robust_makedirs(dir3)

    orig_mkdirs = os.makedirs
    dir4 = str(tmp_path / "d2" / "d3")

    def mymkdirs(path, *args, **kwargs):
        orig_mkdirs(path)
        if path == dir4:  # makedirs is recursive, so raise only for dir4
            raise OSError("can't create directory, it exists already")

    monkeypatch.setattr(os, "makedirs", mymkdirs)
    robust_makedirs(dir4)
    assert os.path.isdir(dir4), "pass on dir created while trying"



# -----------------------------------------------------------------------------
#   Imports
# -----------------------------------------------------------------------------
import threading
import logging

try:
    import IPython.terminal.embed
    HAVE_IPYTHON = True
    HAVE_CODE = False
except Exception:
    HAVE_IPYTHON = False

try:
    import code
    HAVE_CODE = True
except Exception:
    HAVE_CODE = False


logger = logging.getLogger(__name__)


def simple_code_shell(namespace):
    """ basic interactive shell if ipython is not found
        but code can be imported
    """
    try:
        import readline  # noqa: F401
    except Exception:
        print("readline not found")

    shell = code.InteractiveConsole(namespace)
    shell.interact()


class CLI(object):
    """ cli object
    """

    input_func = input

    def __init__(
            self,
            options=None,
            namespace=None,
            quit_func=None,
            use_code=False,  # use code shell if possible
            use_custom=False,  # use custom shell if possible
            ):

        self.use_ipython = self.use_code = self.use_custom = False
        if HAVE_IPYTHON and not (use_code or use_custom):
            self.use_ipython = True
        elif HAVE_CODE and not use_custom:
            self.use_code = True
        else:
            self.use_custom = True
        self._cli_thread = None
        self._options = options
        if namespace is None:
            namespace = dict(__lock=threading.Lock())
        self.namespace = namespace
        if not hasattr(namespace, '__lock'):
            namespace.update(dict(__lock=threading.Lock()))
        self._lock = namespace['__lock']
        self._quit_function = quit_func

    def set_quit_function(self, func):
        self._quit_function = func

    def run(self):
        """ allows to run an ipython shell with the CLI's context vars """
        namespace = self.namespace

        if self.use_ipython:
            logger.debug("CLI using ipython")
            shell = IPython.terminal.embed.InteractiveShellEmbed(
                user_ns=namespace)
            shell()
        elif self.use_code:
            logger.debug("CLI using basic code fallback")
            simple_code_shell(namespace=namespace)
        else:
            logger.debug("CLI using very basic custom fallback")
            self.mini_shell(namespace=namespace)

        if self._quit_function:
            try:
                self._quit_function(self)
            except TypeError:
                logger.warning("using obsolete quit function without argument")
                self._quit_function()

    def mini_shell(self, namespace):
        """ Rather lousy Python shell for debugging in case code can't be
            imported.
            This is probably only the case for some py2exe / pyinstaller
            versions or other incomplete pythons
        """
        while True:
            try:
                cmd_line = self.input_func('--> ')
            except EOFError:
                cmd_line = "q"

            upper_stripped = cmd_line.strip().upper()
            shall_quit = (upper_stripped == 'Q' or upper_stripped == 'QUIT')
            if shall_quit:
                break
            try:
                eval(compile(cmd_line, '<string>', 'single'), namespace)
            except Exception as exc:
                logger.error('ERROR: %r' % exc)

        print("END OF CLI")
        self.write_history()

    def write_history(self, fname=None):
        pass

    def run_as_thread(self, name='cli', daemon=True):
        """ start cli as a thread
            This is needed for Qt Apps, where the GUI must be called
            in the main thread
        """
        self._cli_thread = cli_thread = threading.Thread(
            target=self.run,
            name=name)
        cli_thread.daemon = daemon
        cli_thread.start()


def main():
    pass


if __name__ == '__main__':
    main()
# -----------------------------------------------------------------------------
#   End of file
# -----------------------------------------------------------------------------



import pytest

from mytb.string import str2bool


class AClass(object):
    """ just a dummy type for testing """


def test_str2bool():
    # None should yield None
    assert str2bool(None) is None

    # some legal True strings, ints and floats
    for val in ("yes", "Yes", "TRUE", "y", "T", "1", 1, 2.):
        assert str2bool(val)

    # some legal False strings, ints and floats
    for val in ("nO", "False", "f", "N", "0", 0, 0., ""):
        assert str2bool(val) is False

    # unknwon strings and types shall raise an Exception
    for val in ("a", AClass()):
        with pytest.raises(ValueError):
            str2bool(val)



#!/usr/bin/env python

import csv


class CSVReader(object):
    def __init__(self, fname, encoding='utf8'):
        self.fname = fname
        self.encoding = encoding
        self.fin = None
        self.rdr = None
        self.itr = iter(self)

    def __enter__(self):

        self.fin = open(self.fname, encoding=self.encoding)
        self.rdr = csv.reader(self.fin)
        return self

    def __exit__(self, typ, value, traceback):
        self.fin.close()

    def __iter__(self):
        for row in self.rdr:
            yield row

    def __next__(self):
        return next(self.itr)



import json
import logging

from collections import OrderedDict


logger = logging.getLogger(__name__)


class AnyObj(object):
    """ small object storing arbitrary data in a class
        can be populated via kwargs or via a dict.
        if AnyObj.keys is defined only keys contained in that dict will
        be accepted. others will be ignored
    """

    # TODO: Check whether we shouldn't fail if assigning unknown keys !!!!
    keys = None

    def __init__(self, **kwargs):
        if self.keys is None:
            for akey in kwargs:
                self.__dict__[akey] = kwargs[akey]
        else:
            for akey in kwargs:
                if akey in self.keys:
                    mykey = self.keys.get(akey, akey)
                    self.__dict__[mykey] = kwargs[akey]
        self.convert()

    @staticmethod
    def keys_from_list(key_list):
        return dict((key, key) for key in key_list)

    @classmethod
    def from_dict(cls, a_dict):
        """ populates object from a dict """
        return cls(**a_dict)  # pylint: disable=W0142

    def convert(self):
        """ to be overloaded. permits conversion during object
            creation
        """

    def as_dict(self, recursive=False):
        """ returns vars as a dict """
        adict = vars(self)
        if not recursive:
            return adict
        adict = dict(adict)
        to_update = {}  # don't update a dict while iterating through it
        for key, val in adict.items():
            if hasattr(val, "as_dict"):
                to_update[key] = val.as_dict(recursive=True)
        adict.update(to_update)
        return adict

    def pop(self, attr):
        """ pop / remove an object attribute """
        val = getattr(self, attr)
        delattr(self, attr)
        return val

    def __eq__(self, other):
        return self.as_dict() == other.as_dict()


class VersionObject:
    """ simple wrapper to add type / version info to an object
    """

    o_type = "unknown"
    version = "0.0"

    def __init__(self, data=None, version=None, o_type=None, **hdr_args):
        """ creates a versioned object from data and meta info """
        cls = self.__class__
        if version is None:
            version = cls.version
        self.version = version
        if o_type is None:
            o_type = cls.o_type
        self.o_type = o_type
        self.data = data
        self.hdr_args = OrderedDict(hdr_args.items())

    def as_dict(self):
        """ returns versionobject as dict """
        adict = OrderedDict([("type", self.o_type), ("version", self.version)])
        adict.update(self.hdr_args.items())
        adict["data"] = self.data
        return adict

    @classmethod
    def from_dict(cls, adict):
        """ creates a versioned object from a dict """
        adict = dict(adict)
        o_type = adict.pop("type")
        version = adict.pop("version")
        data = adict.pop("data")
        return cls(data, version, o_type, **adict)

    @classmethod
    def from_json_file(cls, fname):
        logger.debug("FROM_FILE %r", fname)
        with open(fname) as fin:
            data = json.load(fin)
        return cls.from_dict(data)



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.logging.config
#
# Description:  helpers to get fast and quick a decent log configuration
#
# #############################################################################
import os
import sys
import logging
import logging.config
import re
import inspect
# import json
import importlib

import __main__

# from mytb.pprint import pprint
from mytb.importlib import module_exists


def get_default_log_settings(**kwargs):
    """
    creates default logger settings, that are good enough for many cases
    """
    logdir = kwargs.get('log_dir', '.')
    name = kwargs.get('name', 'mytb')
    enable_file_handler = kwargs.get('logtofile')
    handlers = ["console"]
    level = kwargs.get("level", "WARNING")
    cfg = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "verbose": {
                "format": "%(levelname)-8s %(asctime)s %(process)d"
                          + " %(name)-18s:%(lineno)d %(message)s"
            },
            "simple": {
                "format": "%(levelname)-8s %(asctime)s"
                          + " %(name)-18s:%(lineno)d %(message)s",
                "datefmt": "%H:%M:%S",
            },
        },
        "handlers": {
            "console": {
                "level": level,
                "class": "logging.StreamHandler",
                "formatter": "simple"
            },
        },
        "loggers": {
            "": {
                "level": level,
                "handlers": handlers,
            },
        }
    }
    if enable_file_handler:
        handlers["file"] = {
            "level": level,
            "class": "logging.FileHandler",
            "filename": "%s/%s.log" % (logdir, name),
            "formatter": "verbose"
        }
        handlers = ["file"] + handlers
    # print(cfg['handlers']['file'])
    return cfg


get_log_settings = get_default_log_settings


def shall_configure_logging(name=None):
    """ tries to determine whether we should configure
        logging or not
    """
    if name == "__main__":  # simple case. just cfg logging
        return True

    # let's inspect the call stack
    # print("MAIN FILE", getattr(__main__, '__file__', '?'))
    stack = inspect.stack()
    # print(len(stack), "stack entries")
    assert len(stack) >= 3  # at least stack entries should be there
    #  pprint(stack, "STACK")
    record = stack[2]
    #  pprint(record, "\n\nrecord")
    frm, fname, lno, funcname = record[:4]
    # print("fname", fname)
    if fname == getattr(__main__, '__file__', '?'):
        # print("CALLED BY MAIN")
        return True
    dirname, basename = os.path.split(fname)
    # print("DN", dirname, " BN", basename)
    if (os.path.basename(dirname) == 'commands' and
            os.path.splitext(basename)[0] == '__init__'):
        return False
        # return True


def split_config(log_config):
    """
    splits log config into the config name and its arguments

        This is not fully implemented. it will split on ' ' or
            on ':' depending what occurs first
    """
    splitvals = re.split(r'(?:\s+|:)', log_config, 1)
    if len(splitvals) == 2:
        cfg_name, cfg_argstr = splitvals
        cfg_args = {}
        for kv in cfg_argstr.split(','):
            key, val = kv.split('=')
            cfg_args[key] = val
    else:
        cfg_name = log_config
        cfg_args = {}
    return cfg_name, cfg_args


def config_logger(cfg_name, name):
    """ tries to locate a log configuration and apply it.

        Rules:
        1.) split cfg_name into cfg_name and cfg_args
        2.) if cfg_name is a file name, the determine its file type and load it
                accordingly. only python files will handle the passed args
        3.) for cfg_names, that are NOT a file try to import following module
            in following order:
               <cfg_name>
               log_settings.<cfg_name>
               log_settings.<cfg_name>_default
               log_settings.default
               log_settings.default_default
               mytb.logging.configs.<cfg_name>
               mytb.logging.config
            accept the first module, that contains a dict named LOGGING
            or has a callable named get_log_settings(), or a callable
                setup_logging()
    """
    # print("CONFL", cfg_name, name, sys.argv[:1])
    if cfg_name is None:
        if name == "__main__":
            name = os.path.splitext(os.path.basename(sys.argv[0]))[0]
        cfg_name = name

    cfg_name, cfg_args = split_config(cfg_name)
    if 'name' not in cfg_args:
        cfg_args['name'] = name
    # print("N: %r / A: %r" % (cfg_name, cfg_args))

    # can I find a config file
    if os.path.isfile(cfg_name):
        suffix = os.path.splitext(cfg_name)[1].lower()
        # allowed_suffixes = [
        #     '.json', '.yaml', '.ini', '.py', '.pyw', '.pyc', '.pyo' ]
        allowed_suffixes = ['.py', ]
        if suffix in allowed_suffixes:
            print("setup logging from file %s" % cfg_name)
            if suffix == ".py" and cfg_args:
                print("cfg args are %r" % cfg_args)
            raise NotImplementedError("config file handling still to be coded")
            return

    modnames = (  # module names to search
            "{0} log_settings.{0} log_settings.{0}_default "
            "log_settings.default log_settings.default_default "
            "mytb.logging.configs.{0} "
            "mytb.logging.config"
            ).format(cfg_name).split()
    for modname in modnames:
        # print("MN", modname)
        try:
            exists = module_exists(modname)
        except ImportError:
            exists = False
        if exists:
            # print(modname, "exists")
            try:
                mod = importlib.import_module(modname)
            except Exception as exc:
                print("Error when import log configuration", exc)
                raise
            # print("MOD:", mod)
            logcfg = getattr(mod, 'LOGGING', None)
            getcfg = getattr(mod, "get_log_settings", None)
            # print("LOGCFG", logcfg)
            # print("GETCFG", getcfg)
            if logcfg:
                log_dict = getattr(mod, 'LOGGING', None)
            elif getcfg:
                # print("GETCFG", getcfg)
                # print("CFGARGS", cfg_args)
                try:
                    log_dict = getcfg(**cfg_args)
                except Exception:
                    print("Error when retrieving log configuration from ",
                          modname)
                    raise
            else:
                log_dict = None
            if log_dict:
                try:
                    logging.config.dictConfig(log_dict)
                except Exception:
                    print("Error when applying log configuration from ",
                          modname)
                    raise
                return
            setupfunc = getattr(mod, "setup_logging", None)
            if setupfunc:
                setupfunc(**cfg_args)
                return


# def mk_default_logger(cfg_name):
#     cfg = json.loads(DEFAULT_DICT_CFG_STR
#                      % dict(cfg_name=cfg_name, logdir='.'))
#     logging.config.dictConfig(cfg)


def setupLogging(name=None, force_config=False):
    """ if called from a main module or force_config=True
        setup logging according to some rules
        as described in config_logger()
    """
    # print("GETL %s FC=%s SCL %s" %
    #     (name, force_config, shall_configure_logging(name)))
    if force_config or shall_configure_logging(name):
        from mytb.argparse import mk_parser, LONG_LOG_SWITCH
        args = sys.argv[1:]
        parser = mk_parser(add_help=False)
        options, unknown = parser.parse_known_args(args)
        # print("OPTIONS", options)
        log_cfg = getattr(options, LONG_LOG_SWITCH[2:].replace('-', '_'))
        # print("LOG_CFG: %r" % log_cfg)
        config_logger(log_cfg, name)


def getLogger(name=None, force_config=False):
    """ gets a logger (like logging.getLogger()
        and if called from a main module or force_config=True
        setup logging according to some rules
        as described in config_logger()
    """
    setupLogging(name, force_config)
    return logging.getLogger(name)



#!/usr/bin/env python

# ############################################################################
# Copyright  : (C) 2014-2021 by MHComm. All rights reserved
#
# Name       :  mytb.file.find
"""
  Summary    : filefind helpers

__author__    = "Klaus Foerster"
__email__     = "info@mhcomm.fr"

"""
# #############################################################################
import os


# TODO: check if we get this faster / more effient with python3
def file_find_tuple(basedir, relpath=False, dir_filter=None, file_filter=None):
    if relpath is True:
        relpath = basedir
    for bdir, dirs, files in os.walk(basedir):
        if dir_filter:
            dirs[:] = [adir for adir in dirs if dir_filter(adir)]
        dirs[:] = sorted(dirs)
        if relpath:
            bdir = os.path.relpath(bdir, relpath)
        if file_filter:
            files = [fname for fname in files if file_filter(fname)]
        for fname in sorted(files):
            yield (bdir, fname)


def file_find(basedir, relpath=False, dir_filter=None, file_filter=None,
              path_filter=None):
    found_tuples = file_find_tuple(basedir, relpath, dir_filter, file_filter)
    full_paths = (os.path.join(bdir, fname) for bdir, fname in found_tuples)
    if path_filter:
        for full_path in full_paths:
            if path_filter(full_path):
                yield full_path
    else:
        for full_path in full_paths:
            yield full_path



#!/usr/bin/env python

# ############################################################################
# Copyright  : (C) 2014 by MHComm. All rights reserved
#
# module released to mytb. under mytb's license
#
# Name       :  mytb.html.table_reader
"""
  Summary    : reads html tables as for example done by xls exports

__author__    = "Klaus Foerster"
"""
# #############################################################################

import html.parser
import logging

from collections import deque
from collections import OrderedDict


logger = logging.getLogger(__name__)


class TableParser(html.parser.HTMLParser):
    """ simple parser  parsing the first table in an html document """
    S_INIT = 0
    S_TABLE = 1
    S_ROW = 2
    S_COL = 3

    def __init__(self):
        super().__init__()
        self.to_yield = deque()
        self.state = self.S_INIT
        self.tagcnt = 0
        self.rowdata = []
        self.celldata = []
        self.max_cnt = 0

    def handle_starttag(self, tag, attrs):
        state = self.state
        self.tagcnt += 1
        if state == self.S_INIT:
            if tag == 'table':
                self.state = self.S_TABLE
        elif state == self.S_TABLE:
            if tag == 'tr':
                self.state = self.S_ROW
                self.rowdata = []
        elif state == self.S_ROW:
            if tag == 'td':
                self.state = self.S_COL
                self.celldata = []
        elif state == self.S_COL:
            pass

    def handle_data(self, data):
        state = self.state
        if state == self.S_COL:
            self.celldata.append(data)

    def handle_endtag(self, tag):
        state = self.state
        if state == self.S_INIT:
            pass
        elif state == self.S_TABLE:
            if tag == 'table':
                self.state = self.S_INIT
        elif state == self.S_ROW:
            if tag == 'tr':
                self.state = self.S_TABLE
                self.to_yield.append(list(self.rowdata))
        elif state == self.S_COL:
            if tag == 'td':
                self.state = self.S_ROW
                self.rowdata.append(''.join(self.celldata))


class TableReader(object):
    def __init__(self, fin, encoding='utf-8'):
        """
            :param fin: input file or iterator yielding
                    characters from an object
        """
        self._from_fname = False
        if isinstance(fin, str):
            fin = open(fin, encoding=encoding)
            self._from_fname = True

        self.fin = fin
        self.parser = TableParser()
        self.cnt = 0

    def rows(self):
        parser = self.parser
        to_yield = parser.to_yield
        try:
            for line in self.fin:
                self.cnt += 1
                parser.feed(line)

                while len(to_yield):
                    rslt = to_yield.popleft()
                    yield rslt
        except Exception:
            pass
        finally:
            if self._from_fname:
                self.fin.close()

    def __iter__(self):
        return self.rows()


class TableDictReader(TableReader):
    """ Table reader returning dicts
    """
    def rows(self, cls=OrderedDict):
        """
        :param :cls dict class or factory to construct
            an iterated object from header row / data row
        """
        rows = TableReader.rows(self)
        head_row = next(rows)
        logger.debug("HEADROW %r", head_row)
        for row in rows:
            if len(row) != len(head_row):
                row = row + [None] * (len(head_row) - len(row))
            yield cls(zip(head_row, row))


def main():
    pass


if __name__ == '__main__':
    main()
# -----------------------------------------------------------------------------
#   End of file
# -----------------------------------------------------------------------------



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.pprint
#
# Description:  lazy pretty print
#
# #############################################################################


def pprint(obj, head=None, indent=1):
    """
    """
    from pprint import PrettyPrinter
    pp = PrettyPrinter(indent=indent)
    if head:
        print(head)
    pp.pprint(obj)



import pytz
import re
import tzlocal

from datetime import datetime
from datetime import timedelta

import dateutil.parser


# datetime objct for beginning of epoch
T_EPOCH = datetime(1970, 1, 1, tzinfo=pytz.utc)
DEFAULT = object()  # singleton, for args with default values


class DateTimeError(Exception):
    """ custom exception """


class DateTime(object):
    single_delta = r'(?:\s*([+-]\d+(?:\.\d*)?)(?:\s*([shMdw])?)\s*)'
    single_delta = r'(?:\s*([+-]\d+(?:\.\d*)?)\s*([shMdw]?)\s*)'
    # attempt to handle comma separated list of deltas
    # multi_delta = r'^%s(?:,%s)*$' % (single_delta, single_delta)
    delta_rex = re.compile('^' + single_delta + '$')

    delta_units = {
        's': (0, 1),
        'M': (0, 60),
        'h': (0, 3600),
        'd': (1, 0),
        'w': (7, 0),
        '': (1, 0),  # default unit = days
        }

    @classmethod
    def strptimedelta(cls, deltastr, info=None, raise_on_error=True):
        """ parses a date time string and returns a datetime timedelta object
             Supported Formats:
                 '+-<num><unit>'
                 where unit =
                     s for seconds
                     h for hours
                     M for minutes
                     d for days
                     w for weeks
                     default = days
        """
        # not implemented so far
        #             and rounding (use by strptime) =
        #                 d for days
        #                 default no rounding

        #     """
        # TODO: think about using dateutil.parser.relativedelta
        rslt = datetime.now(pytz.utc)
        fields = (val.strip() for val in deltastr.split(','))
        delta_rex = cls.delta_rex
        for field in fields:
            match = delta_rex.match(field)
            if not match:
                raise DateTimeError("can't parse %r as delta" % field)
            value, unit = match.groups()
            value = float(value)
            days, seconds = cls.delta_units[unit]
            rslt += timedelta(days * value, seconds * value)

        return rslt

    @classmethod
    def strptime(cls, datestr=None, fmt=None, tzinfo=DEFAULT):
        """ parses a date time string and returns a date time object
            Supported Formats:
                - formats as supported by dateutil.parser
                - None, '', 0, '0' and 'now' -> datetime.now()
                - if fmt is passed same as datetime.strptime

            :param datestr: date string to be passed
            :param fmt: if passedm then use datetime's normal strptime
                        BUT add a time zone info
            :param tzinfo: if no tz info is specified in the string, then
                this param decides which time zone shall be used.
                DEFAULT: use local time zone
                None: return naive time zone object
                other: use other time zone

        """
        #    NOT IMPLEMENTED SO FAR
        #        - delta format with +-num units[rounding],
        #        where unit =
        #            s for seconds
        #            M for minutes
        #            h for hours
        #            d for days
        #            w for weeks
        #        and rounding =
        #            d for days
        #            default no rounding

        tzinfo = tzinfo if tzinfo is not DEFAULT else tzlocal.get_localzone()
        if fmt:
            rslt = datetime.strptime(datestr, fmt)
        else:
            if isinstance(datestr, (int, float)):
                datestr = str(datestr)
            datestr = datestr.strip() if datestr else datestr
            if datestr in (None, '', '0', 'now'):
                return datetime.now(tzinfo)
            if datestr[:1] in "+-" or ',' in datestr:
                return cls.strptimedelta(datestr, tzinfo)

            rslt = dateutil.parser.parse(datestr)

        if rslt.tzinfo is None and tzinfo:
            rslt = tzinfo.localize(rslt)
        return rslt

    @classmethod
    def parse_range(cls, rangestr=None, default_from='-1d', default_to='now'):
        """ parses a time range string
            a time range string is a comma separated list of a start time
            and a end time
        """

        if rangestr is None:
            from_str = default_from
            to_str = default_to
        else:
            from_str, to_str = [v.strip() for v in rangestr.split(',', 1)]
            from_str = from_str if from_str else default_from
            to_str = to_str if to_str else default_to
        t_from = cls.strptime(from_str)
        t_to = cls.strptime(to_str)
        return t_from, t_to


class Time(DateTime):
    @classmethod
    def strptime(cls, datestr):
        pass


class Date(DateTime):
    @classmethod
    def strptime(cls, datestr):
        pass


def fname_to_time(fname, use_ctime=False, use_mtime=False, tz=None):
    """ extracts date time from an fname
        examples of supported formats:
            "fnameYYYYMMDD"            just a date
            "fnameYYYY-MM-DD"          date with separators
            "fnameYYYYMMDD_HHmmss"     date and time
            "fnameYYYYMMDD-HHmmss"     date and time
            "fnameYYYYMMDD-HH-mm-ss"   date and time
            "fnameYYYYMMDD-ssssssssss" date and time(in seconds since epoche)

        :param fname: file name to parse
        :param use_ctime: if file name contains no string use file's ctime
        :param use_mtime: if file name contains no string use file's mtime

    """


def to_timestamp(t):
    """ convert a datetime object to seconds since epoch """
    return (t - T_EPOCH).total_seconds()



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.commands
#
"""
    Entry point for mytb commands
"""
#
# #############################################################################
# python std modules
import argparse
import sys

# third party modules
import minibelt

# TODO: add autodiscovery
commands = [(
        "check_ci_cfg", "mytb.gitlab.check_ci_cfg.main",
        "checks config of a gitlab ci cfg",
        )]

cmd_dict = dict((entry[0], entry[1:]) for entry in commands)


def show_commands(options=None):
    """ shows list of available commands """
    print("available mytb commands:")
    for entry in commands:
        cmd, mod, helptext = entry
        print("%18s: %s" % (cmd, helptext))


def mk_parser():
    parser = argparse.ArgumentParser(description="entrypoint for mytb scripts")
    parser.add_argument(
            "cmd", nargs=argparse.REMAINDER,
            help="command to call")
    return parser


def main():
    """ the main function """
    args = sys.argv[1:]
    parser = mk_parser()
    options = parser.parse_args(args)
    cmd = options.cmd
    if not cmd:
        show_commands(options)
        return
    cmd_name = cmd[0]
    if cmd_name not in cmd_dict:
        print("ERROR: command %s is unnown:\n" % cmd_name)
        show_commands(options)
        sys.exit(255)

    mod, _help_text = cmd_dict[cmd_name]
    func = minibelt.import_from_path(mod)
    sys.argv = args
    func()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017-2021 by Teledomic.eu All rights reserved
#
# Name:         mytb.importlib
#
# Description:  helper for locating / importing modules
#
# #############################################################################
import importlib


def module_exists(modulename):
    """ tries to find out whether a module exists """
    import pkgutil
    moduleparts = modulename.split('.')
    mod = ""
    for part in moduleparts:
        mod += '.' + part if mod else part
        if pkgutil.find_loader(mod) is None:
            return False
    return True


def import_obj(obj_str):
    """
    handles input string as module.object

    and imports the module and returns the object
    """
    if ':' in obj_str:
        modname, objname = obj_str.split(':')
    else:
        modname, objname = obj_str.rsplit('.', 1)
    mod = importlib.import_module(modname)
    return getattr(mod, objname)


def import_if_mod_exists(mod_or_func_str):
    """ attempts to import a module or an object if the
        nodule exists

        returns None if module is not existing and should
        raise an exception if there's another problem
        during the module's import

        :param mod_or_func_str: if str contains ":", then the last
                                part is considered to be an object within
                                the module


    """
    mod_name = mod_or_func_str.split(":")[0]
    get_obj = ":" in mod_or_func_str

    rslt = None
    try:
        if get_obj:
            rslt = import_obj(mod_or_func_str)
        else:
            rslt = importlib.import_module(mod_or_func_str)
    except ModuleNotFoundError as exc:
        msg = str(exc)
        if "'%s'" % mod_name not in msg:
            raise

    return rslt



def get_log_settings(basedir='', basename='log', level='DEBUG', **kwargs):
    log_settings = {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'verbose': {
                'format': '%(levelname)-8s %(asctime)s %(name)s:%(lineno)d '
                          '%(message)s'
            },
            'simple': {
                'format': '%(levelname)s %(module)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': level,
                'class': 'logging.StreamHandler',
                'formatter': 'verbose'
            },
        },
        'loggers': {
            # root loggers
            '': {
                'level': level,
                'handlers': ['console'],
                'propagate': False,
            },
        }
    }
    return log_settings



# first party modules
import yaml


class Pattern(object):
    def __init__(self, value):
        if type(value) is dict:
            self.value = yaml.dump(value)
            print(repr(self.value))
        else:
            self.value = value


pass_patterns = [
    Pattern(  # valid cache entry
        dict(cache=dict(key="123", paths=[".", "other"]))
    ),
    Pattern(  # valid stages and before entry
        dict(
            stages=["one", "two"],
            before_script=["one", "two"],
            )
        ),
    Pattern({  # rather complete cfg
        'cache': dict(key="1234", paths=[".", "tst"]),
        'stages': ["one", "two"],
        'before_script': ["one", "two"],
        'setup': dict(
            stage="one",
            script=["cmd1", "cmd2"],
            ),
        'test2': dict(
            stage="one",
            script=["cmd1", "cmd2"],
            tags=["tag1"],
            ),
        }),
]

fail_patterns = [
    Pattern(  # invalid yaml syntax
        " val1:\nval2:\n"
        ),
    Pattern(  # cache has unknown keys
        dict(cache=dict(unknownkey=1))
        ),
    Pattern(  # stages must be list
        dict(stages=3)
        ),
    Pattern(  # stages must be list
        dict(stages=dict(name='bad'))
        ),
    Pattern(  # before_script must be list
        dict(before_script=3)
        ),
    # cache.paths must be list
    Pattern(
        dict(cache=dict(paths=3))
        ),
]



import importlib
import logging
import os
import re


logger = logging.getLogger(__name__)


def get_modules(paths=None, exclude_test=False, report=None):
    """ finds modules in a list of paths """
    suffixes = ['.py', '.pyc', '.pyo']
    if report is None:
        report = {}
        error_modules = report['error_modules'] = []
    for pathdir, mod_prefix in paths:
        for basedir, dirs, files in os.walk(pathdir):
            if exclude_test and 'test' in dirs:  # skip test modules
                dirs.remove('test')
            relpath = os.path.relpath(basedir, pathdir)
            if relpath == '.':
                mod_path = mod_prefix
            else:
                mod_path = mod_prefix + '.' + re.sub(r'[\\/]', '.', relpath)

            # keep only python files
            fnames = (
                fname for fname in files
                if os.path.splitext(fname)[1] in suffixes)

            # get and unify module names
            fnames = set(re.sub(r'\.py[co]?$', '', fname) for fname in fnames)

            # traverse modules in alphabetical order
            for fname in sorted(list(fnames)):
                if fname != "__init__":
                    mod_name = mod_path + '.' + fname
                else:
                    mod_name = mod_path
                try:
                    mod = importlib.import_module(mod_name)
                except Exception:
                    logger.warning("failed importing %r", mod_name)
                    error_modules.append(mod_name)
                yield(mod)


def mk_modname(pm_name):
    """ normalizes a filename or module name
    normalize test name
    examples: "tests/modules/tst_01.py" -> "tests.modules.tst_01"
    examples: "./tests/modules/tst_01.py" -> "tests.modules.tst_01"
    examples: "tests\\modules\\tst_01.py" -> "tests.modules.tst_01"
    """
    mod_name = pm_name
    if os.path.isfile(mod_name):
        mod_name = os.path.splitext(os.path.normpath(pm_name))[0]
        mod_name = mod_name.replace('/', '.')
        mod_name = mod_name.replace('\\', '.')
    return mod_name



#!/usr/bin/env python

# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.argparse
#
# Description:  helper for some common argparser flags
#
# #############################################################################


import os
import sys
import argparse

from mytb.string import str2bool

LONG_LOG_SWITCH = "--log-config"
SHORT_LOG_SWITCH = "-L"
LOG_CONF_ENV_VAR = "MYTB_LOG_CONFIG"


def mk_parser(description="", app_name=None, add_help=True):
    """ creates a parser with some default switches """
    log_config = os.environ.get(
            LOG_CONF_ENV_VAR, "mytb.logging.configs.console:level=WARNING")
    # print("LOG_CONFIG = %r" % log_config)

    parser = argparse.ArgumentParser(
        description=description,
        add_help=add_help)
    if not SHORT_LOG_SWITCH:
        parser.add_argument(
            LONG_LOG_SWITCH,
            default=log_config,
            help="name of config module (and params)")
    else:
        parser.add_argument(
            LONG_LOG_SWITCH, '-L',
            default=log_config,
            help=("name of config module (and params) "
                  "default=%(default)s. Can "
                  "be overridden with env var {}".format(LOG_CONF_ENV_VAR))
            )

    if not app_name:
        app_name = os.path.splitext(os.path.basename(sys.argv[0]))[0]

    parser.set_defaults(app_name=app_name)

    return parser


def add_bool_argument(parser, *args, **kwargs):
    """
    adds a parser for parsing params as true or false
    """
    mykwargs = dict(
        type=str2bool,
        const=True,
        nargs="?"
        )
    mykwargs.update(kwargs)
    parser.add_argument(*args, **mykwargs)



# python std modules

# third party modules
from setuptools import find_packages
from setuptools import setup

long_description = """small modules and tools useful for many projects
"""

setup(name="mytb",
      version="0.1.1",
      description="my toolbox for everyday python projects",
      long_description=long_description,
      long_description_content_type="text/x-rst",
      classifiers=[
            "Development Status :: 3 - Alpha",
      ],
      keywords="toolbox development",
      url="https://github.com/feenes",
      author="Teledomic",
      author_email="info@teledomic.eu",
      license="MIT",
      packages=find_packages(),
      scripts=[],
      entry_points={
          "console_scripts": [
            "mytb = mytb.commands:main",
            ]
      },
      project_urls={
        "Homepage": "https://github.com/feenes/mytb",
        "Documentation": "https://github.com/feenes/mytb",
        "Source": "https://github.com/feenes/mytb",
        "SayThanks": "https://github.com/feenes",
        "Funding": "https://donate.pypi.org",
        "Tracker": "https://github.com/feenes/mytb/issues",
      },
      install_requires=[
        "minibelt",
        "unidecode",
        ],
      extras_require=dict(
        minimal=[],
        all=[
            "dateutils",
            "ddt",
            "pytz",
            "pyyaml",
            "tzlocal",
            "unidecode",
            ],
        date=[
            "dateutils",
            "pytz",
            "tzlocal",
            ],
        gitlab=[
            "ddt",
            "pyyaml",
            ],
        ),
      python_requires=">=3.6, <4",
      setup_requires=["pytest-runner"],
      tests_require=["pytest", "pytest-tempdir"],
      zip_safe=False,
      include_package_data=True,
      )



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.tests.ipc
#
# Description:  ipc test case
#
# #############################################################################
# import unittest

# from mytb.ipc.mutex import Mutex
#
# class IPCTestCase(unittest.TestCase):
#     def test_get_rls_mutex(self):
#         """ can acquire and release a mutex """
#         mutex = Mutex("toto")
#         status = mutex.status
#         print("MST", status)
#         self.assertFalse(mutex.status)
#         mutex.acquire(timeout=1)
#         self.assertTrue(mutex.status)
#         mutex.release()
#         self.assertFalse(mutex.status)
#     def test_ctx_get_rls_mutex(self):
#         """ can acquire and release with context manager """
#         with Mutex("toto", 1) as mutex:
#             pass
#             #self.assertTrue(mutex.status)
#         #self.assertFalse(mutex.status)
#         # can acquire mutex without timeout
#         mutex = Mutex("toto", 1)
#
#
# if __name__ == "__main__":
#     import sys
#     args = sys.argv[1:]
#     if args and args[0] == "test":
#         print("TST")
#         suite = unittest.TestSuite()
#         suite.addTest(IPCTestCase("test_get_rls_mutex"))
#         runner = unittest.TextTestRunner(verbosity=2, stream=sys.stdout)
#         runner.run(suite)



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.os
#
# Description:  os helpers
#
# #############################################################################
import os
import sys


_RELEASE = None  # well, OS version just has to be determined once


def is_windows(min_release=None, max_release=None):
    """ True if OS is WINDOWS  (between min / max_release) """
    global _RELEASE

    if sys.platform != 'win32':
        return False

    if _RELEASE is None:
        _RELEASE = sys.platform.release()

    if min_release is not None and int(_RELEASE) < min_release:
        return False

    if max_release is not None and int(_RELEASE) > max_release:
        return False

    return True


def is_linux():
    """ True if OS is linux """
    return sys.platform.startswith('linux')


def is_posix():
    """ True if OS is posix """
    return os.name == 'posix'


def is_macos_x():
    """ True if OS is mac """
    return sys.platform == 'darwin'


def is_cygwin():
    """ True if OS is cygwin """
    return sys.platform == 'cygwin'


def is_win_7():
    """ True if OS is windows 7 """
    return is_windows() and _RELEASE == '7'


def is_win_8():
    """ True if OS is windows 8 """
    return is_windows() and _RELEASE == '8'


def is_win_10():
    """ True if OS is windows 10 """
    return is_windows() and _RELEASE == '10'



#!/usr/bin/env python
# #############################################################################
# Copyright : (C) 2017 by Teledomic.eu All rights reserved
#
# Name:         mytb.tempfile
#
# Description:  helpers for temp files
#
# #############################################################################
import os
import tempfile


def mktemp_fname(suffix='', prefix='tmp', dir=None):
    """ creates a temporary filename and closes all open handles
        in order to be nice with OSes like Windows
    """
    fdesc, fname = tempfile.mkstemp(suffix=suffix, prefix=prefix, dir=dir)
    os.close(fdesc)
    return fname
